# coding: utf-8
from lxml.etree import tostring
from spyne.util.xml import get_object_as_xml
from spyne import Fault

from kinder.core.unit.models import Unit
from kinder.core.dict.models import UnitKind

from .period import period
from .interfaces import ModelIndex, CalculatedIndex
from .types import Item, Root

from . import strategies
from . import logger

UNIT_FIELDS = ['id', 'name', 'full_name']


def strategies_map():
    """
    Возвращает маппинг стратегий из модуля.

    Благодаря тому что список классов берется из модуля
    для заведения нового показателя достаточно добавить класс по
    аналогии с остальными.
    """

    index = (ModelIndex, CalculatedIndex)
    return {
        name: class_
        for name, class_ in strategies.__dict__.iteritems()
        # Если попытаться проверить issubclass не для класса
        # то упадет TypeError и весь цикл вместе с ним
        if isinstance(class_, type) and issubclass(class_, index)}


class UnitIterator(object):
    """Итератор по учреждениям региона.

    Возвращает список учреждений, добавляя
    информацию о том к какому МО оно принадлежит
    """
    def __init__(self, region_id):
        self.region_id = region_id

    def __iter__(self):
        mos = Unit.objects.filter(
            kind_id=UnitKind.MO,
            parent_id=self.region_id)
        for mo in mos:
            descendants = (
                mo.get_descendants()
                .filter(kind_id=UnitKind.DOU)
                .values(*UNIT_FIELDS)
                .exclude(id=mo.id))
            mo = {field: getattr(mo, field) for field in UNIT_FIELDS}
            for child in descendants:
                child.update({'mo': mo})
                yield child


class Analytic(object):
    """Класс отвечающий запуск для каждого учреждения стратегии подсчета."""
    def __init__(self, strategy, extra_param):
        self.strategy = strategy()
        # Дополнительный параметр для подсчета показателя
        # к примеру для подсчета показателя по желаемым учреждениям
        # необходимо передать позицию ЖУ в очереди
        self.strategy.extra_param = extra_param

    def region_data(self, period):
        """Возвращает плоский список учреждений с результатом подсчета."""
        region_ids = Unit.objects.filter(
            kind_id=UnitKind.REGION).values_list('id', flat=True)
        for region_id in region_ids:
            units = UnitIterator(region_id)
            for unit in units:
                count = self.strategy.calculate(unit, period)
                logger.debug(
                    'Unit id: {}, result: {}'.format(unit['id'], count))
                unit.update({'count': count})
                yield Item(
                    count=count,
                    unit_id=unit['id'],
                    unit_name=unit['full_name'],
                    unit_shortname=unit['name'],
                    control_unit=unit['mo']['full_name'],
                    control_unit_id=unit['mo']['id'],
                )

    def process(self, period):
        return Root(units=tuple(self.region_data(period)))


def get_content(**kwargs):
    """Получает и сериализует данные в указанный формат."""

    # Получение из запроса периода, тип отчета и допольнительного параметра
    index_period = period(
        kwargs['period_type'],
        start=kwargs.get('date_from'),
        end=kwargs.get('date_to'),
    )
    report_strategy = strategies_map()[kwargs['report_code']]
    extra_param = kwargs.get('param')

    logger.debug(
        'Requested "{}" report, extra params: {}'.format(
            kwargs['report_code'], extra_param))
    logger.debug('Period({}, {} - {})'.format(
        index_period.type, index_period.start, index_period.end))

    root = Analytic(report_strategy, extra_param).process(index_period)

    return tostring(
        get_object_as_xml(root), pretty_print=True, encoding='unicode')
